; PROGRAM: IL_EX.ASM (Example program with subroutines to drive the
; LCD screen of the ILM-216 LCD module)
; Scott Edwards Electronics, Inc. 1998

; NOTE: This program was originally written in Parallax assembler. 
; A program was used to translate it to Microchip dialect. This 
; accounts for some odd formatting, etc. Make sure that 
; your assembler is set as follows: 
;  RADIX = DECIMAL
;  CASE SENSITIVE = ON
;  PROCESSOR = 16F84. 
; This program was test assembled and verified with MPASM for Windows, 
; version 2.01.08. We cannot vouch for backward compatibility with 
; earlier MPASMs, as Microchip's documentations says there 
; are differences. 

; This program provides a very basic starting point for programmers
; working on alternative applications for the ILM-216 LCD module
; from Scott Edwards Electronics Inc. (www.seetron.com). It is not
; meant as a finished application or as a tutorial on LCDs or PICs.
; It's 'Hello World' program to get you off to a quick start with
; driving the LCD portion of the module.

; Programming note: This example uses the older TRIS method of setting
; up I/O lines, and all I/Os not actually used in the demo are set to
; output. There is no 'upward compatibility' problem with this appraoch
; because this code is specifically for the 16F84 used in the ILM-216.
; Use of TRIS will provoke four harmless "warnings" during assembly. 

; ==SPECIAL FUNCTION REGISTERS
pc	equ	2	; Program counter. 
status	equ	3	; Status register. 
ra	equ	5	; Port RA.
rb	equ	6	; Port RB.
#define	Z	2	; (Z)ero flag. 

; ==VARIABLES
temp    equ     H'c'    ; Reusable temporary storage.
LCDout  equ     H'd'    ; Data or instruction to write to LCD.
LCDin   equ     H'e'    ; Data read from the LCD.
hafMs   equ     H'f'    ; Number of 0.5-ms units to delay in Pause.
rbCopy  equ     H'10'   ; Copy of port rb used during LCD writes.

; ==EQUATES
; Control lines to LCD module.
#define RS      ra,1    ; 0 = instruction       1 = data.
#define RW      ra,2    ; 0 = write             1 = read.
#define E       ra,3    ; 0 = disable           1 = enable.
; Other I/O lines
#define BELL    ra,4    ; "Bell" output or GP I/O.
#define BL      rb,4    ; Backlight control; 1= ON.
; Special numbers
#define BUS_OUT 0       ; TRIS setting for output to 4-bit LCD bus.
#define BUS_IN  B'1111' ; TRIS setting for input from 4-bit LCD bus

; ==LCD instructions (not all instructions listed are used in this demo)
; Instructions must be sent using the wr_inst subroutine. For more
; info on the instructions, see the KS0066 docs (or HD44780 docs).
; A quick-reference table of instructions appears in S.E.E.'s OEM
; manual for the ILM-216.
#define CLS     1       ; Clear-screen and home cursor.
#define HOME    2       ; Move cursor to 1st char of 1st line (home)
#define LINE1   H'80'   ; Address of 1st character of 1st line.
#define LINE2   H'c0'   ; Address of 1st character of 2nd line.
#define SCRLFT  H'18'   ; Scroll entire screen 1 char left.
#define SCRRHT  H'1c'   ; Scroll entire screen 1 char right.
#define ULCURS  H'e'    ; Turn on underline cursor.
#define BBCURS  H'd'    ; Turn on blinking-block cursor.
#define NOCURS  H'c'    ; Turn cursor off.
#define BLANK   8       ; Blank the display (but keep the data).
#define RESTORE H'c'    ; Restore 'blanked' display (w/cursor off).
#define DDRAM   H'80'   ; Point to beginning of display (DD) RAM.
#define CGRAM   H'40'   ; Point to beginning of character-gen (CG) RAM.

; ==QUICK PRIMER ON LCD SCREEN LAYOUT
; This information is covered in excruciating detail in the KS0066 docs,
; but here's a quick rundown. The LCD controller has a total of 80 bytes
; of RAM. When you display text on the LCD, you are actually writing
; data to display RAM (DD RAM). Since there is more RAM than screen
; locations (80 versus 32), not all data written to RAM shows up on the
; screen. So, in order to print just to the screen and not to invisible,
; off-screen RAM, you must sometimes move the cursor (or DD RAM pointer,
; to be technical about it). Here's a map of the 2x16 screen and the
; DD RAM instructions (in hex) to place the cursor at those locations:
;======================================================================
; position:    0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
; line 1:     80 81 82 83 84 85 86 87 88 89 8A 8B 8C 8D 8E 8F
; line 2:     C0 C1 C2 C3 C4 C5 C6 C7 C8 C9 CA CB CC CD CE CF
;======================================================================
; Bottom line: when you print to the end of line 1, you must send 0C0h
; as an instruction to move the cursor to the beginning of line 2. If
; you don't, the next 24 bytes sent will be invisible.

        org 0
        goto start      ; Jump over subroutines to beginning of demo.
        org 4

; ==========================
;        DEMO DATA
; ==========================
; This table contains the data that will be written to the LCD.
; The routine that writes to the display looks for the 0s at the
; ends of the lines as end-of-string markers.

string  addwf pc,1
        retlw 'H'
        retlw 'e'
        retlw 'l'
        retlw 'l'
        retlw 'o'
        retlw H'20'
        retlw 'W'
        retlw 'o'
        retlw 'r'
        retlw 'l'
        retlw 'd'
        retlw '!'
        retlw 0

        retlw 'I'
        retlw 'L'
        retlw 'M'
        retlw '-'
        retlw '2'
        retlw '1'
        retlw '6'
        retlw 'L'
        retlw 0

; ==========================
;    LCD SUBROUTINES
; ==========================

; Produce an approximate 10-ms delay used during initialization of
; LCD before busy flag can be read. Can also be used to produce a
; variable delay in 0.5-ms units up to 127ms. Place the delay
; value in hafMs and call Pause.

delay10
        movlw D'20'
        movwf hafMs
Pause
Pause_load      movlw D'248'    ; Set up for 1-ms loop.
        movwf temp
Pause_loop
        nop             ; Do nothing.
        decfsz temp,1           ; Do more nothing.
        goto Pause_loop
        decfsz hafMs,1          ; Repeat hiB times.
        goto Pause_load
        retlw 0

; Pulse the LCD enable line high for a couple of microseconds.
blip_e
        bsf E
        goto $+1
        goto $+1
        bcf E
        retlw 0

; Initialize the LCD in 4-bit mode. Set for standard operation of the
; 2x16 LCD--normal 5x8 pixel font, 2-line operation, cursor movement
; from left to right, display on.
init_LCD
        movlw B'11'             ; First init step
        movwf rb
        call blip_e
        call delay10
        call blip_e             ; second...
        call delay10
        call blip_e             ; third...
        movlw B'10'             ; set 4-bit interface.
        movwf rb
        call blip_e
        call delay10
        movlw B'101000'                 ; Function set.
        movwf LCDout
        call wr_inst
        call ck_busy
        movlw B'1100'                   ; Display ON.
        movwf LCDout
        call wr_inst
        call ck_busy
        movlw B'110'                    ; Print left-to-right.
        movwf LCDout
        call wr_inst
        call ck_busy
        movlw CLS                       ; Clear the screen.
        movwf LCDout
        call wr_inst
        call ck_busy
        retlw 0

; Write instructions or data to the LCD, depending on entry point to the
; subroutine. Data or instruction to be written must be in LCDout.
; Data is transferred in two 4-bit chunks.
; Note that this routine copies the latch states of port RB and combines
; those bits with the data to be written to the LCD via RB0-RB3. This
; preserves the state of the backlight and RB5-RB7.

wr_inst bcf RS
	goto skip1
wr_data bsf RS
skip1   bcf RW
        movf rb,0
        andlw B'11110000'
        movwf rbCopy
        swapf LCDout,0
        andlw B'1111'
        iorwf rbCopy,0
        movwf rb
        call blip_e
        movf LCDout,0
        andlw B'1111'
        iorwf rbCopy,0
        movwf rb
        call blip_e
        retlw 0

; Read data back from the LCD. Calling rd_addr gets you the current DD RAM
; address of the LCD memory pointer (i.e., the cursor location, but
; expressed as an address). Bit 7 of the address data is the busy bit,
; which must be 0 before the LCD can accept new data.
; Calling rd_data gets the ASCII character at the current DD RAM
; address/cursor location.

rd_addr bcf RS
	goto skip2
rd_data bsf RS
skip2   bsf RW                          ; Put LCD into read mode.
        movlw BUS_IN                    ; Bus to input.
        tris rb
        bsf E
        nop
        swapf rb,0
        movwf LCDin
        movlw B'11110000'
        andwf LCDin,1
        bcf E
        nop
        bsf E
        nop
        movf rb,0
        bcf E
        andlw B'1111'
        iorwf LCDin,1
        movlw BUS_OUT
        tris rb
        bcf RW
        retlw 0

; Read the busy bit and loop until 0 (not busy).

ck_busy call rd_addr
        btfsc LCDin,7
        goto ck_busy
        retlw 0

; ==========================
;    DEMO PROGRAM
; ==========================

; This program uses the previously defined subroutines to write a message
; to the LCD screen. The message is stored in a lookup table at the
; beginning of this program.

start   clrf ra
        movlw 0                 ; Set ra to all outputs.
        tris ra
        movlw BUS_OUT           ; Set rb to all outputs.
        tris rb
        call init_LCD
        bsf BL                  ; Turn on the backlight

; ==Note that the backlight should be turned on after initialization,
; because init routine does not preserve previous state of other RB bits.

        clrf temp               ; Start at beginning of string table (0).

start_line1
        call ck_busy            ; Test for busy before writes.
        movf temp,0             ; Point to current char of string (w).
        incf temp,1             ; Aim temp at next char of string.
        call string             ; Get char in w.
        movwf LCDout            ; Transfer to LCDout.
        movf LCDout,1           ; Check for 0 in LCDout.
        btfsc status,Z
        goto start_end1
        call wr_data            ; ..else write to display.
        goto start_line1        ; Repeat.


start_end1
        movlw LINE2             ; Set cursor to line 2.
        movwf LCDout
        call wr_inst            ; write as instruction.


start_line2
        call ck_busy            ; Same as line 1 above..
        movf temp,0
        incf temp,1
        call string             ; Get char in w.
        movwf LCDout            ; Transfer to LCDout.
        movf LCDout,1
        btfsc status,Z
        goto start_end
        call wr_data
        goto start_line2
start_end
        goto $                  ; Stop program (endless loop).
        end
